from pathlib import Path
import re
import zipfile

import ase
import ase.io

from franken.datasets.registry import DATASET_REGISTRY, BaseRegisteredDataset
from franken.utils.file_utils import download_file


@DATASET_REGISTRY.register("water")
class WaterRegisteredDataset(BaseRegisteredDataset):
    relative_paths = {
        "water": {
            "train": "water/ML_AB_dataset_1.xyz",
            "val": "water/ML_AB_dataset_2-val.xyz",
        },
    }
    zip_file_names = ["ML_AB_dataset_1", "ML_AB_dataset_2", "ML_AB_128h2o_validation"]

    @classmethod
    def get_path(
        cls, name: str, split: str, base_path: Path | None, download: bool = True
    ):
        if base_path is None:
            raise KeyError(None)
        relative_path = cls.relative_paths[name][split]
        path = base_path / relative_path
        if not path.is_file() and download:
            cls.download(base_path)
        if path.is_file():
            return path
        else:
            raise ValueError(f"Dataset not found at '{path.resolve()}'")

    @classmethod
    def download(cls, base_path: Path):
        water_base_path = base_path / "water"
        water_base_path.mkdir(exist_ok=True, parents=True)

        # NOTE: cannot check MD5 here since it changes at every download. As a dumb fallback we check the file-size.
        download_file(
            url="https://zenodo.org/api/records/10723405/files-archive",
            filename=water_base_path / "data.zip",
            desc="Downloading water dataset",
            expected_size=35866571,
        )
        # Extract from zip and convert VASP -> XYZ format
        with zipfile.ZipFile(water_base_path / "data.zip", mode="r") as zf:
            for file_name in cls.zip_file_names:
                with zf.open(file_name, "r") as fh:
                    vasp_data = fh.read().decode("utf-8")
                    xyz_data = vasp_mlff_to_xyz(vasp_data)
                with open(water_base_path / f"{file_name}.xyz", "w") as fh:
                    fh.write(xyz_data)
                # Sanity check
                traj = ase.io.read(water_base_path / f"{file_name}.xyz", index=":")
                assert isinstance(traj, list)
                for i, atoms in enumerate(traj):
                    atoms.get_potential_energy()
                    atoms.get_forces()
        # Split a validation set from dataset-2
        dataset = ase.io.read(
            water_base_path / "ML_AB_dataset_2.xyz", index=":", format="extxyz"
        )
        assert isinstance(dataset, list)
        dataset_no_overlap = dataset[473:]
        ase.io.write(water_base_path / "ML_AB_dataset_2-val.xyz", dataset_no_overlap)
        # Cleanup
        (water_base_path / "data.zip").unlink()


def vasp_mlff_to_xyz_oneconfig(data):
    # Parse sections using regular expressions
    num_atoms = int(re.search(r"The number of atoms\s*[-=]+\s*(\d+)", data).group(1))
    energy = float(
        re.search(r"Total energy \(eV\)\s*[-=]+\s*([-+]?\d*\.\d+|\d+)", data).group(1)
    )

    # Extract lattice vectors
    lattice_match = re.search(
        r"Primitive lattice vectors \(ang.\)\s*[-=]+\s*([\d\s.-]+)", data
    )
    lattice_lines = lattice_match.group(1).strip().split("\n")
    lattice = [line.split() for line in lattice_lines]

    # Flatten and format lattice as a string for XYZ format
    lattice_flat = " ".join([" ".join(line) for line in lattice])

    # Extract atomic positions
    positions_match = re.search(
        r"Atomic positions \(ang.\)\s*[-=]+\s*([\d\s.-]+)", data
    )
    positions_lines = positions_match.group(1).strip().split("\n")
    positions = [line.split() for line in positions_lines]

    # Extract forces
    forces_match = re.search(r"Forces \(eV ang.\^-1\)\s*[-=]+\s*([\d\s.-]+)", data)
    forces_lines = forces_match.group(1).strip().split("\n")
    forces = [line.split() for line in forces_lines]

    # Extract stress tensor (two lines) without separators
    stress_match_1 = re.search(
        r"Stress \(kbar\)\s*[-=]+\s*XX YY ZZ\s*[-=]+\s*([\d\s.-]+)", data
    )
    stress_match_2 = re.search(r"XY YZ ZX\s*[-=]+\s*([\d\s.-]+)", data)

    # Ensure we only capture numerical values and not separator lines
    stress_values_1 = (
        stress_match_1.group(1).strip().split()[:3]
    )  # Take first three values for XX YY ZZ
    stress_values_2 = (
        stress_match_2.group(1).strip().split()[:3]
    )  # Take first three values for XY YZ ZX
    xx, yy, zz = stress_values_1
    xy, yz, zx = stress_values_2

    # Combine the two stress components into a single list
    # stress_tensor = stress_values_1 + stress_values_2
    # stress_tensor = ' '.join(stress_tensor)  # Convert to a single string
    stress_tensor = f"{xx} {xy} {zx} {xy} {yy} {yz} {zx} {yz} {zz}"

    # Create the extended XYZ content for this configuration
    xyz_content = []
    xyz_content.append(f"{num_atoms}")
    xyz_content.append(
        f'Lattice="{lattice_flat}" Properties=species:S:1:pos:R:3:forces:R:3 energy={energy} stress="{stress_tensor}"'
    )

    # Atom types (order them according to the positions provided)
    atom_type_lines = (
        re.search(r"Atom types and atom numbers\s*[-=]+\s*([\w\s\d]+)", data)
        .group(1)
        .strip()
        .split("\n")
    )
    atom_types = []
    for line in atom_type_lines:
        element, count = line.split()
        atom_types.extend([element] * int(count))

    # Add each atom's data line by line
    for idx, (position, force) in enumerate(zip(positions, forces)):
        element = atom_types[idx]
        px, py, pz = position
        fx, fy, fz = force
        xyz_content.append(f"{element} {px} {py} {pz} {fx} {fy} {fz}")

    return "\n".join(xyz_content)


def vasp_mlff_to_xyz(data):
    # Split the data by configurations using "Configuration num." as the delimiter
    configurations = re.split(r"Configuration num\.\s*\d+", data)
    xyz_all = []

    # Process each configuration if it is not empty
    for config in configurations:
        config = config.strip()
        if config:  # Only parse if the configuration is not empty
            try:
                xyz_all.append(vasp_mlff_to_xyz_oneconfig(config))
            except AttributeError:
                pass  # some errors are expected.

    # Join all configurations with a newline
    return "\n".join(xyz_all)


if __name__ == "__main__":
    WaterRegisteredDataset.download(Path(__file__).parent.parent)



import numpy as np
import torch

from franken.data.base import Target
from franken.metrics.base import BaseMetric
from franken.metrics.registry import registry
from franken.utils import distributed


__all__ = [
    "EnergyMAE",
    "EnergyRMSE",
    "ForcesMAE",
    "ForcesMAESpecies",
    "ForcesRMSE",
    "ForcesRMSESpecies",
    "ForcesCosineSimilarity",
    "is_pareto_efficient",
]


class EnergyMAE(BaseMetric):
    def __init__(self, device: torch.device, dtype: torch.dtype = torch.float32):
        units = {
            "inputs": "eV",
            "outputs": "meV/atom",
        }
        super().__init__("energy_MAE", device, dtype, units)

    def update(self, predictions: Target, targets: Target) -> None:
        if targets.forces is None:
            raise NotImplementedError(
                "At the moment, target's forces are required to get the number of atoms in the configuration."
            )
        num_atoms = targets.forces.shape[-2]
        num_samples = 1
        if targets.energy.ndim > 0:
            num_samples = targets.energy.shape[0]

        error = 1000 * torch.abs(targets.energy - predictions.energy) / num_atoms

        self.buffer_add(error, num_samples=num_samples)


class EnergyRMSE(BaseMetric):
    def __init__(self, device: torch.device, dtype: torch.dtype = torch.float32):
        units = {
            "inputs": "eV",
            "outputs": "meV/atom",
        }
        super().__init__("energy_RMSE", device, dtype, units)

    def update(self, predictions: Target, targets: Target) -> None:
        if targets.forces is None:
            raise NotImplementedError(
                "At the moment, target's forces are required to get the number of atoms in the configuration."
            )
        num_atoms = targets.forces.shape[-2]
        num_samples = 1
        if targets.energy.ndim > 0:
            num_samples = targets.energy.shape[0]

        error = torch.square((targets.energy - predictions.energy) / num_atoms)

        self.buffer_add(error, num_samples=num_samples)

    def compute(self, reset: bool = True) -> torch.Tensor:
        if self.buffer is None:
            raise ValueError(
                f"Cannot compute value for metric '{self.name}' "
                "because it was never updated."
            )
        distributed.all_sum(self.buffer)
        distributed.all_sum(self.samples_counter)
        error = self.buffer / self.samples_counter
        # square-root and fix units
        error = torch.sqrt(error) * 1000
        if reset:
            self.reset()
        return error


class ForcesMAE(BaseMetric):
    def __init__(self, device: torch.device, dtype: torch.dtype = torch.float32):
        units = {
            "inputs": "eV/ang",
            "outputs": "meV/ang",
        }
        super().__init__("forces_MAE", device, dtype, units)

    def update(self, predictions: Target, targets: Target) -> None:
        if targets.forces is None or predictions.forces is None:
            raise AttributeError("Forces must be specified to compute the MAE.")
        num_samples = 1
        if targets.forces.ndim > 2:
            num_samples = targets.forces.shape[0]
        elif targets.forces.ndim < 2:
            raise ValueError("Forces must be a 2D tensor or a batch of 2D tensors.")

        error = 1000 * torch.abs(targets.forces - predictions.forces)
        error = error.mean(dim=(-1, -2))  # Average over atoms and components

        self.buffer_add(error, num_samples=num_samples)


class ForcesMAESpecies(BaseMetric):
    """
    Returns force MAE computed for each species.
    """

    Z_MAX = 90  # upper bound

    def __init__(self, device: torch.device, dtype: torch.dtype = torch.float32):
        units = {
            "inputs": "eV/ang",
            "outputs": "meV/ang",
        }
        super().__init__(
            name="forces_MAE_species",
            device=device,
            dtype=dtype,
            units=units,
            requires_species=True,
        )

        # buffers will be initialized later once we know n_models
        self.buffer = None
        self.samples_counter = torch.zeros(self.Z_MAX + 1, device=device, dtype=dtype)

    def reset(self) -> None:
        if self.buffer is not None:
            self.buffer.zero_()
        self.samples_counter.zero_()

    def update(
        self,
        predictions: Target,
        targets: Target,
        atomic_numbers: torch.Tensor,
    ) -> None:

        # ---- assertions on species ----
        assert atomic_numbers.ndim == 1
        assert atomic_numbers.shape[0] == targets.forces.shape[-2]
        assert atomic_numbers.max() <= self.Z_MAX

        if targets.forces is None or predictions.forces is None:
            raise AttributeError("Forces must be specified to compute the MAE.")
        # |ΔF| in eV/Å, averaged over Cartesian components
        # shapes:
        #   single model: (N,)
        #   ensemble:     (M, N)
        error = torch.abs(targets.forces - predictions.forces).mean(dim=-1)
        if error.ndim == 1:
            error = error.unsqueeze(0)  # (1, N)

        n_models = error.shape[0]

        # lazy buffer initialization
        if self.buffer is None:
            self.buffer = torch.zeros(
                self.Z_MAX + 1,
                n_models,
                device=self.device,
                dtype=self.dtype,
            )

        species = torch.unique(atomic_numbers)

        # accumulate per species
        for z in species:
            z_int = int(z)
            mask = atomic_numbers == z  # (N,)

            # sum over atoms, keep models
            # (M, N_z) → (M,)
            self.buffer[z_int] += error[:, mask].sum(dim=1)

            # count atoms (same for all models)
            self.samples_counter[z_int] += mask.sum()

    def compute(self, reset: bool = True) -> torch.Tensor:
        if self.buffer is None:
            raise ValueError(
                f"Cannot compute value for metric '{self.name}' "
                "because it was never updated."
            )

        # sync across ranks
        distributed.all_sum(self.buffer)
        distributed.all_sum(self.samples_counter)

        # buffer shape: (Z, M) → transpose to (M, Z)
        buffer = self.buffer.transpose(0, 1)  # (M, Z)

        # MAE per model, per species
        mae = torch.zeros_like(buffer)

        mask = self.samples_counter > 0
        mae[:, mask] = buffer[:, mask] / self.samples_counter[mask]

        # unit conversion: eV/Å → meV/Å
        mae = mae * 1000

        # store average across present species at index 0
        species_mask = mask.clone()
        species_mask[0] = False
        if species_mask.any():
            mae[:, 0] = mae[:, species_mask].mean(dim=1)

        if reset:
            self.reset()

        return mae


class ForcesRMSESpecies(BaseMetric):
    """
    Returns force RMSE computed for each species.
    """

    Z_MAX = 90  # upper bound

    def __init__(self, device: torch.device, dtype: torch.dtype = torch.float32):
        units = {
            "inputs": "eV/ang",
            "outputs": "meV/ang",
        }
        super().__init__(
            name="forces_RMSE_species",
            device=device,
            dtype=dtype,
            units=units,
            requires_species=True,
        )

        # buffers will be initialized later once we know n_models
        self.buffer = None
        self.samples_counter = torch.zeros(self.Z_MAX + 1, device=device, dtype=dtype)

    def reset(self) -> None:
        if self.buffer is not None:
            self.buffer.zero_()
        self.samples_counter.zero_()

    def update(
        self,
        predictions: Target,
        targets: Target,
        atomic_numbers: torch.Tensor,
    ) -> None:

        # ---- assertions on species ----
        assert atomic_numbers.ndim == 1
        assert atomic_numbers.shape[0] == targets.forces.shape[-2]
        assert atomic_numbers.max() <= self.Z_MAX

        if targets.forces is None or predictions.forces is None:
            raise AttributeError("Forces must be specified to compute the RMSE.")
        # ΔF^2 in (eV/Å)^2, averaged over Cartesian components
        # shapes:
        #   single model: (N,)
        #   ensemble:     (M, N)
        error = torch.square(targets.forces - predictions.forces).mean(dim=-1)
        if error.ndim == 1:
            error = error.unsqueeze(0)  # (1, N)

        n_models = error.shape[0]

        # lazy buffer initialization
        if self.buffer is None:
            self.buffer = torch.zeros(
                self.Z_MAX + 1,
                n_models,
                device=self.device,
                dtype=self.dtype,
            )

        species = torch.unique(atomic_numbers)

        # accumulate per species
        for z in species:
            z_int = int(z)
            mask = atomic_numbers == z  # (N,)

            # sum over atoms, keep models
            # (M, N_z) -> (M,)
            self.buffer[z_int] += error[:, mask].sum(dim=1)

            # count atoms (same for all models)
            self.samples_counter[z_int] += mask.sum()

    def compute(self, reset: bool = True) -> torch.Tensor:
        if self.buffer is None:
            raise ValueError(
                f"Cannot compute value for metric '{self.name}' "
                "because it was never updated."
            )

        # sync across ranks
        distributed.all_sum(self.buffer)
        distributed.all_sum(self.samples_counter)

        # buffer shape: (Z, M) -> transpose to (M, Z)
        buffer = self.buffer.transpose(0, 1)  # (M, Z)

        # mean squared error per model, per species
        mse = torch.zeros_like(buffer)

        mask = self.samples_counter > 0
        mse[:, mask] = buffer[:, mask] / self.samples_counter[mask]

        # RMSE and unit conversion: eV/Å -> meV/Å
        rmse = torch.sqrt(mse) * 1000

        # store average across present species at index 0
        species_mask = mask.clone()
        species_mask[0] = False
        if species_mask.any():
            rmse[:, 0] = rmse[:, species_mask].mean(dim=1)

        if reset:
            self.reset()

        return rmse


class ForcesRMSE(BaseMetric):
    def __init__(self, device: torch.device, dtype: torch.dtype = torch.float32):
        units = {
            "inputs": "eV/ang",
            "outputs": "meV/ang",
        }
        super().__init__("forces_RMSE", device, dtype, units)

    def update(self, predictions: Target, targets: Target) -> None:
        if targets.forces is None or predictions.forces is None:
            raise AttributeError("Forces must be specified to compute the MAE.")
        num_samples = 1
        if targets.forces.ndim > 2:
            num_samples = targets.forces.shape[0]
        elif targets.forces.ndim < 2:
            raise ValueError("Forces must be a 2D tensor or a batch of 2D tensors.")

        error = torch.square(targets.forces - predictions.forces)
        error = error.mean(dim=(-1, -2))  # Average over atoms and components

        self.buffer_add(error, num_samples=num_samples)

    def compute(self, reset: bool = True) -> torch.Tensor:
        if self.buffer is None:
            raise ValueError(
                f"Cannot compute value for metric '{self.name}' "
                "because it was never updated."
            )
        distributed.all_sum(self.buffer)
        distributed.all_sum(self.samples_counter)
        error = self.buffer / self.samples_counter
        # square-root and fix units
        error = torch.sqrt(error) * 1000
        if reset:
            self.reset()
        return error


class ForcesRMSE2(BaseMetric):
    """Average of RMSE along individual structures"""

    def __init__(self, device: torch.device, dtype: torch.dtype = torch.float32):
        units = {
            "inputs": "eV/ang",
            "outputs": "meV/ang",
        }
        super().__init__("forces_RMSE", device, dtype, units)

    def update(self, predictions: Target, targets: Target) -> None:
        if targets.forces is None or predictions.forces is None:
            raise AttributeError("Forces must be specified to compute the MAE.")
        num_samples = 1
        if targets.forces.ndim > 2:
            num_samples = targets.forces.shape[0]
        elif targets.forces.ndim < 2:
            raise ValueError("Forces must be a 2D tensor or a batch of 2D tensors.")

        error = torch.square(targets.forces - predictions.forces)
        error = error.mean(dim=(-1, -2))  # Average over atoms and components
        error = torch.sqrt(error) * 1000
        self.buffer_add(error, num_samples=num_samples)


class ForcesCosineSimilarity(BaseMetric):
    def __init__(self, device: torch.device, dtype: torch.dtype = torch.float32):
        units = {
            "inputs": "eV/ang",
            "outputs": None,
        }
        super().__init__("forces_cosim", device, dtype, units)

    def update(
        self,
        predictions: Target,
        targets: Target,
    ) -> None:
        num_samples = 1
        assert targets.forces is not None
        assert predictions.forces is not None
        if targets.forces.ndim > 2:
            num_samples = targets.forces.shape[0]
        elif targets.forces.ndim < 2:
            raise ValueError("Forces must be a 2D tensor or a batch of 2D tensors.")

        cos_similarity = torch.nn.functional.cosine_similarity(
            predictions.forces, targets.forces, dim=-1
        )
        cos_similarity = cos_similarity.mean(dim=-1)
        self.buffer_add(cos_similarity, num_samples=num_samples)


def is_pareto_efficient(costs):
    """
    Find the pareto-efficient points
    :param costs: An (n_points, n_costs) array
    :return: A (n_points, ) boolean array, indicating whether each point is Pareto efficient
    """
    is_efficient = np.ones(costs.shape[0], dtype=bool)
    for i, c in enumerate(costs):
        if is_efficient[i]:
            is_efficient[is_efficient] = np.any(
                costs[is_efficient] < c, axis=1
            )  # Keep any point with a lower cost
            is_efficient[i] = True  # And keep self
    return is_efficient


registry.register("energy_MAE", EnergyMAE)
registry.register("energy_RMSE", EnergyRMSE)
registry.register("forces_MAE", ForcesMAE)
registry.register("forces_RMSE", ForcesRMSE)
registry.register("forces_RMSE2", ForcesRMSE2)
registry.register("forces_cosim", ForcesCosineSimilarity)
registry.register("forces_MAE_species", ForcesMAESpecies)
registry.register("forces_RMSE_species", ForcesRMSESpecies)



import pytest
import torch
import importlib.util
import e3nn

from franken.backbones import REGISTRY
from franken.backbones.utils import load_checkpoint
from franken.config import BackboneConfig, GaussianRFConfig
from franken.data import BaseAtomsDataset
from franken.datasets.registry import DATASET_REGISTRY
from franken.rf.model import FrankenPotential

from packaging.version import Version

from franken.utils.misc import no_jit

# Check availability of backbones
HAS_MACE = importlib.util.find_spec("mace") is not None
HAS_SEVENN = importlib.util.find_spec("sevenn") is not None
HAS_FAIRCHEM = importlib.util.find_spec("fairchem") is not None

# Build parametrized model list with skip marks when deps are missing
models = []
for name in REGISTRY.keys():
    kind = REGISTRY[name]["kind"]
    marks = []
    if (kind == "mace" and not HAS_MACE) or (kind == "sevenn" and not HAS_SEVENN) or (kind == "fairchem" and not HAS_FAIRCHEM):
        marks.append(pytest.mark.skip(reason=f"{kind} not installed"))
    elif "SchNet" in name:
        marks.append(pytest.mark.xfail(reason="Fails in CI due to unknown reasons", strict=False))
    elif kind == "mace":
        marks.append(pytest.mark.xfail(Version(e3nn.__version__) >= Version("0.5.5"), reason="Known incompatibility", strict=True))
    elif kind == "sevenn":
        marks.append(pytest.mark.xfail(Version(e3nn.__version__) < Version("0.5.0"), reason="Known incompatibility", strict=True))
    models.append(pytest.param(name, marks=marks))

@pytest.mark.parametrize("model_name", models)
def test_backbone_loading(model_name):
    registry_entry = REGISTRY[model_name]
    gnn_config = BackboneConfig.from_ckpt(
        {
            "family": registry_entry["kind"],
            "path_or_id": model_name,
            "interaction_block": 2,
        }
    )
    load_checkpoint(gnn_config)


@pytest.mark.parametrize("model_name", models)
def test_descriptors(model_name):
    registry_entry = REGISTRY[model_name]
    gnn_config = BackboneConfig.from_ckpt(
        {
            "family": registry_entry["kind"],
            "path_or_id": model_name,
            "interaction_block": 2,
        }
    )
    bbone = load_checkpoint(gnn_config)
    # Get a random data sample
    data_path = DATASET_REGISTRY.get_path("test", "train", None, False)
    dataset = BaseAtomsDataset.from_path(
        data_path=data_path,
        split="train",
        gnn_config=gnn_config,
    )
    data, _ = dataset[0]  # type: ignore
    expected_fdim = bbone.feature_dim()
    features = bbone.descriptors(data)
    assert features.shape[1] == expected_fdim


@pytest.mark.parametrize("model_name", models)
def test_force_maps(model_name):
    from franken.backbones.wrappers.common_patches import patch_e3nn

    patch_e3nn()
    registry_entry = REGISTRY[model_name]
    gnn_config = BackboneConfig.from_ckpt(
        {
            "family": registry_entry["kind"],
            "path_or_id": model_name,
            "interaction_block": 2,
        }
    )
    # Get a random data sample
    data_path = DATASET_REGISTRY.get_path("test", "train", None, False)
    dataset = BaseAtomsDataset.from_path(
        data_path=data_path,
        split="train",
        gnn_config=gnn_config,
    )
    device = "cuda:0" if torch.cuda.is_available() else "cpu"
    # initialize model
    model = FrankenPotential(
        gnn_config=gnn_config,
        rf_config=GaussianRFConfig(num_random_features=128, length_scale=1.0),
    )
    model = model.to(device)
    data, _ = dataset[0]  # type: ignore
    data = data.to(device)
    with torch.no_grad(), no_jit():
        # Need to call this multiple times to make sure test passes!
        emap, fmap = model.grad_feature_map(data)
        emap, fmap = model.grad_feature_map(data)
        emap, fmap = model.grad_feature_map(data)



import importlib.resources
import json
import logging
import os
from pathlib import Path
from packaging.version import Version

import requests
import torch

from franken.config import BackboneConfig, asdict_with_classvar
from franken.utils import distributed
from franken.utils.file_utils import download_file


logger = logging.getLogger("franken")


def load_model_registry():
    model_registry_text = (
        importlib.resources.files("franken.backbones")
        .joinpath("registry.json")
        .read_text()
    )
    model_registry = json.loads(model_registry_text)
    return model_registry


class CacheDir:
    directory: Path | None = None

    @staticmethod
    def initialize(cache_dir: Path | str | None = None):
        if CacheDir.is_initialized():
            logger.warning(
                f"Cache directory already initialized at {CacheDir.directory}. Reinitializing."
            )
        # Default cache location: ~/.franken
        default_cache = Path.home() / ".franken"
        if cache_dir is None:
            env_cache_dir = os.environ.get("FRANKEN_CACHE_DIR", None)
            if env_cache_dir is None:
                logger.info(f"Initializing default cache directory at {default_cache}")
                cache_dir = default_cache
            else:
                logger.info(
                    f"Initializing cache directory from $FRANKEN_CACHE_DIR {env_cache_dir}"
                )
                cache_dir = env_cache_dir
        else:
            logger.info(f"Initializing custom cache directory {cache_dir}")
        CacheDir.directory = Path(cache_dir)

        # Ensure the directory exists
        if not CacheDir.directory.exists():
            CacheDir.directory.mkdir(parents=True, exist_ok=True)
            logger.info(f"Created cache directory at: {CacheDir.directory}")

    @staticmethod
    def get() -> Path:
        if not CacheDir.is_initialized():
            CacheDir.initialize()
        assert CacheDir.directory is not None
        return CacheDir.directory

    @staticmethod
    def is_initialized() -> bool:
        return CacheDir.directory is not None


def make_summary(cache_dir: str | None = None):
    """Function to print available models, grouped and sorted by backbone kind."""
    if cache_dir is not None:
        CacheDir.initialize(cache_dir=cache_dir)

    registry = load_model_registry()
    ckpt_dir = CacheDir.get() / "gnn_checkpoints"

    local_models = []
    remote_models = []

    # Collect implemented backbones
    for model, info in registry.items():
        if not info.get("implemented", False):
            continue
        kind = info["kind"]
        local_path = ckpt_dir / info["local"]
        (local_models if local_path.is_file() else remote_models).append((model, kind))

    # ---- Group → Sort → Flatten ----
    def group_and_sort(models):
        groups = {}
        for model, kind in models:
            groups.setdefault(kind, []).append(model)
        # sort kinds and sort models inside groups
        return {kind: groups[kind] for kind in groups.keys()}

    local_groups = group_and_sort(local_models)
    remote_groups = group_and_sort(remote_models)

    summary = ""

    # ---------------- LOCAL ----------------
    if local_groups:
        summary += f"{'DOWNLOADED MODELS':^80}\n"
        summary += f"{'(' + str(ckpt_dir) + ')':-^80}\n"
        for kind, models in local_groups.items():
            summary += f"* {kind.upper()}\n"
            for model in models:
                summary += f"{model}\n"
            summary += "\n"  # spacing between groups

    # ---------------- REMOTE ----------------
    summary += f"{'AVAILABLE MODELS':-^80}\n"
    for kind, models in remote_groups.items():
        summary += f"* {kind.upper()}\n"
        for model in models:
            summary += f"{model}\n"

    summary += "-" * 80
    return summary


def get_checkpoint_path(backbone_path_or_id: str) -> Path:
    """Fetches the path of a given backbone. If the backbone is not present, it will be downloaded.

    The backbone can be either specified directly via its file-system path,
    then this function is a thin wrapper -- or it can be specified via its
    ID in the model registry. Then this function takes care of finding the
    correct model path and potentially downloading the backbone from the internet.

    Args:
        backbone_path_or_id (str): file-system path to the backbone
            or the backbone's ID as per the model registry.

    Returns:
        Path: Path to the model on disk

    See Also:
        You can use the command :code:`franken.backbones list` from the command-line
        to find out which backbone IDs are supported out-of-the-box.
    """
    registry = load_model_registry()
    gnn_checkpoints_dir = CacheDir.get() / "gnn_checkpoints"

    if backbone_path_or_id not in registry.keys():
        if not os.path.isfile(backbone_path_or_id):
            raise FileNotFoundError(
                f"GNN Backbone path '{backbone_path_or_id}' does not exist. "
                f"You should either provide an existing backbone path or a backbone ID "
                f"from the registry of available backbones: \n{make_summary()}"
            )
        return Path(backbone_path_or_id)
    else:
        backbone_info = registry[backbone_path_or_id]
        ckpt_path = gnn_checkpoints_dir / backbone_info["local"]
        # Download checkpoint being aware of multiprocessing
        if distributed.get_rank() != 0:
            distributed.barrier()
        else:
            if not ckpt_path.exists():
                download_checkpoint(backbone_path_or_id)
            distributed.barrier()
    return ckpt_path


def download_checkpoint(gnn_backbone_id: str, cache_dir: str | None = None) -> None:
    """Download the model if it's not already present locally."""
    registry = load_model_registry()
    if cache_dir is not None:
        CacheDir.initialize(cache_dir=cache_dir)
    ckpt_dir = CacheDir.get() / "gnn_checkpoints"

    if gnn_backbone_id not in registry.keys():
        raise NameError(
            f"Unknown {gnn_backbone_id} GNN backbone, the current available backbones are\n{make_summary()}"
        )

    if not registry[gnn_backbone_id]["implemented"]:
        raise NotImplementedError(
            f"The model {gnn_backbone_id} is not implemented in franken yet."
        )

    local_path = ckpt_dir / registry[gnn_backbone_id]["local"]
    remote_path = registry[gnn_backbone_id]["remote"]

    if local_path.is_file():
        logger.info(
            f"Model already exists locally at {local_path}. No download needed."
        )
        return

    local_path.parent.mkdir(parents=True, exist_ok=True)
    logger.info(f"Downloading model from {remote_path} to {local_path}")
    try:
        download_file(url=remote_path, filename=local_path, desc="Downloading model")
    except requests.RequestException as e:
        logger.error(f"Download failed. {e}")
        raise e


def load_checkpoint(gnn_config: BackboneConfig) -> torch.nn.Module:
    gnn_config_dict = asdict_with_classvar(gnn_config)
    gnn_backbone_id = gnn_config_dict.pop("path_or_id")
    backbone_family = gnn_config_dict.pop("family")
    ckpt_path = get_checkpoint_path(gnn_backbone_id)
    err_msg = f"franken wasn't able to load {gnn_backbone_id}. Is {backbone_family} installed?"
    if backbone_family == "fairchem":
        try:
            from franken.backbones.wrappers.fairchem_schnet import FrankenSchNetWrap
        except ImportError as import_err:
            fairchem_importable = True
            is_fairchem_gt2 = False
            try:
                import fairchem.core

                is_fairchem_gt2 = Version(fairchem.core.__version__) >= Version("2")
                print(f"{is_fairchem_gt2=}")
            except:  # noqa: E722
                fairchem_importable = False
            err_msg = f"franken wasn't able to load {gnn_backbone_id}. "
            if fairchem_importable:
                if is_fairchem_gt2:
                    err_msg += (
                        "Fairchem version < 2 is required. Please see "
                        "https://github.com/facebookresearch/fairchem?tab=readme-ov-file#looking-for-fairchem-v1-models-and-code "
                        "to know more"
                    )
                else:
                    err_msg += import_err.msg
            else:
                err_msg += (
                    f"Please install fairchem version 1.\nBase error: {import_err.msg}"
                )
            logger.error(err_msg, exc_info=import_err)
            raise
        return FrankenSchNetWrap.load_from_checkpoint(
            str(ckpt_path), gnn_backbone_id=gnn_backbone_id, **gnn_config_dict
        )
    elif backbone_family == "mace":
        try:
            from franken.backbones.wrappers.mace_wrap import FrankenMACE
        except ImportError as import_err:
            logger.error(err_msg, exc_info=import_err)
            raise
        return FrankenMACE.load_from_checkpoint(
            str(ckpt_path),
            gnn_backbone_id=gnn_backbone_id,
            map_location="cpu",
            **gnn_config_dict,
        )
    elif backbone_family == "sevenn":
        try:
            from franken.backbones.wrappers.sevenn import FrankenSevenn
        except ImportError as import_err:
            logger.error(err_msg, exc_info=import_err)
            raise
        return FrankenSevenn.load_from_checkpoint(
            ckpt_path, gnn_backbone_id=gnn_backbone_id, **gnn_config_dict
        )
    else:
        raise ValueError(f"Unknown backbone family {backbone_family}")



"""
Script to automatically generate the MACE part of the model registry using the list of released MACE models.
Some urls are available also inside the mace package: https://github.com/ACEsuit/mace/blob/main/mace/calculators/foundations_models.py
"""

import os
import json

# retrieve mace_mp models from mace
# from mace.calculators.foundations_models import mace_mp_urls

mace_urls = {
    "mace_mp/small": "https://github.com/ACEsuit/mace-mp/releases/download/mace_mp_0/2023-12-10-mace-128-L0_energy_epoch-249.model",
    "mace_mp/medium": "https://github.com/ACEsuit/mace-mp/releases/download/mace_mp_0/2023-12-03-mace-128-L1_epoch-199.model",
    "mace_mp/large": "https://github.com/ACEsuit/mace-mp/releases/download/mace_mp_0/MACE_MPtrj_2022.9.model",
    "mace_mp/small-0b": "https://github.com/ACEsuit/mace-mp/releases/download/mace_mp_0b/mace_agnesi_small.model",
    "mace_mp/medium-0b": "https://github.com/ACEsuit/mace-mp/releases/download/mace_mp_0b/mace_agnesi_medium.model",
    "mace_mp/small-0b2": "https://github.com/ACEsuit/mace-mp/releases/download/mace_mp_0b2/mace-small-density-agnesi-stress.model",
    "mace_mp/medium-0b2": "https://github.com/ACEsuit/mace-mp/releases/download/mace_mp_0b2/mace-medium-density-agnesi-stress.model",
    "mace_mp/large-0b2": "https://github.com/ACEsuit/mace-mp/releases/download/mace_mp_0b2/mace-large-density-agnesi-stress.model",
    "mace_mp/medium-0b3": "https://github.com/ACEsuit/mace-mp/releases/download/mace_mp_0b3/mace-mp-0b3-medium.model",
    "mace_mpa/medium-0": "https://github.com/ACEsuit/mace-mp/releases/download/mace_mpa_0/mace-mpa-0-medium.model",
    "mace_omat/small-0": "https://github.com/ACEsuit/mace-mp/releases/download/mace_omat_0/mace-omat-0-small.model",
    "mace_omat/medium-0": "https://github.com/ACEsuit/mace-mp/releases/download/mace_omat_0/mace-omat-0-medium.model",
    "mace_matpes/pbe-0": "https://github.com/ACEsuit/mace-foundations/releases/download/mace_matpes_0/MACE-matpes-pbe-omat-ft.model",
    "mace_matpes/r2scan-0": "https://github.com/ACEsuit/mace-foundations/releases/download/mace_matpes_0/MACE-matpes-r2scan-omat-ft.model",
    "mace_mh/0": "https://github.com/ACEsuit/mace-foundations/releases/download/mace_mh_1/mace-mh-0.model",
    "mace_mh/1": "https://github.com/ACEsuit/mace-foundations/releases/download/mace_mh_1/mace-mh-1.model",
    "mace_omol/0_1024": "https://github.com/ACEsuit/mace-foundations/releases/download/mace_omol_0/MACE-omol-0-extra-large-1024.model",
    "mace_omol/0_4M": "https://github.com/ACEsuit/mace-foundations/releases/download/mace_omol_0/mace-omol-0-extra-large-4M.model",
    "mace_off/small": "https://github.com/ACEsuit/mace-off/blob/main/mace_off23/MACE-OFF23_small.model?raw=true",
    "mace_off/medium": "https://github.com/ACEsuit/mace-off/raw/main/mace_off23/MACE-OFF23_medium.model?raw=true",
    "mace_off/medium24": "https://github.com/ACEsuit/mace-off/blob/main/mace_off24/MACE-OFF24_medium.model?raw=true",
    "mace_off/large": "https://github.com/ACEsuit/mace-off/blob/main/mace_off23/MACE-OFF23_large.model?raw=true",
}

if __name__ == "__main__":
    # update registry
    with open("registry.json", "r") as f:
        model_registry = json.load(f)

    for mace_id, mace_url in mace_urls.items():
        print(mace_id)
        model_registry[f"{mace_id}"] = {
            "remote": mace_url,
            "local": "mace/" + os.path.basename(mace_url),
            "implemented": True,
            "kind": "mace",
        }

    with open("registry.json", "w") as f:
        json.dump(model_registry, f, indent=2)



import os
import argparse
from typing import Optional
import torch
from e3nn.util import jit

from franken.data.base import Configuration
from franken.rf.model import FrankenPotential


@jit.compile_mode("script")
class LammpsFrankenCalculator(torch.nn.Module):
    def __init__(
        self,
        franken_model: FrankenPotential,
    ):
        """Initialize LAMMPS Calculator

        Args:
            franken_model (FrankenPotential): The base franken model used in this MD calculator

        Note:
            The backbone underlying the franken model must be a MACE model. This is because we
            are re-using the LAMMPS interface developed by the MACE authors.
        """
        super().__init__()

        self.model = franken_model
        self.register_buffer("atomic_numbers", self.model.gnn.atomic_numbers)
        self.register_buffer("r_max", self.model.gnn.r_max)
        self.register_buffer("num_interactions", self.model.gnn.num_interactions)
        # this attribute is used for dtype detection in LAMMPS-MACE.
        # See: https://github.com/ACEsuit/lammps/blob/mace/src/ML-MACE/pair_mace.cpp#314
        self.model.node_embedding = self.model.gnn.node_embedding

        for param in self.model.parameters():
            param.requires_grad = False

    def forward(
        self,
        data: dict[str, torch.Tensor],
        local_or_ghost: torch.Tensor,
        compute_virials: bool = False,
    ) -> dict[str, torch.Tensor | None]:
        """Compute energies and forces of a given configuration.

        This module is meant to be used in conjunction with LAMMPS,
        and this function should not be called directly. The format of
        the input data is designed to work with the MACE-LAMMPS fork.

        Warning:
            Stresses and virials are not supported by franken. Since they
            are required to be set by LAMMPS, this function sets them to tensors
            of the appropriate shape filled with zeros. Make sure that
            the chosen MD method does not depend on these quantities.
        """
        # node_attrs is a one-hot representation of the atom types. atom_nums should be the actual atomic numbers!
        # we rely on correct sorting. This is the same as in MACE.
        atom_nums = self.atomic_numbers[torch.argmax(data["node_attrs"], dim=1)]

        franken_data = Configuration(
            atom_pos=data["positions"].double(),
            atomic_numbers=atom_nums,
            natoms=torch.tensor(
                len(atom_nums), dtype=torch.int32, device=atom_nums.device
            ).view(1),
            node_attrs=data["node_attrs"].double(),
            edge_index=data["edge_index"],
            shifts=data["shifts"],
            unit_shifts=data["unit_shifts"],
        )
        energy, forces = self.model(franken_data)  # type: ignore
        # Kokkos doesn't like total_energy_local and only looks at node_energy.
        # We hack around this:
        node_energy = energy.repeat(len(atom_nums)).div(len(atom_nums))
        virials: Optional[torch.Tensor] = None
        if compute_virials:
            virials = torch.zeros((1, 3, 3), dtype=forces.dtype, device=forces.device)
        return {
            "total_energy_local": energy,
            "node_energy": node_energy,
            "forces": forces,
            "virials": virials,
        }

    @staticmethod
    def create_lammps_model(model_path: str, rf_weight_id: int | None) -> str:
        """Compile a franken model into a LAMMPS calculator

        Args:
            model_path (str):
                path to the franken model checkpoint.
            rf_weight_id (int | None):
                ID of the random feature weights. Can generally be left to ``None`` unless
                the checkpoint contains multiple trained models.

        Returns:
            str: the path where the LAMMPS-compatible model was saved to.
        """
        franken_model = FrankenPotential.load(
            model_path,
            map_location=torch.device("cuda" if torch.cuda.is_available() else "cpu"),
            rf_weight_id=rf_weight_id,
        )
        # NOTE:
        # Kokkos is hardcoded to double and will silently corrupt data if the model
        # does not use dtype double.
        franken_model = franken_model.double().to("cpu")
        lammps_model = LammpsFrankenCalculator(franken_model)
        lammps_model_compiled = jit.compile(lammps_model)

        save_path = f"{os.path.splitext(model_path)[0]}-lammps.pt"
        print(f"Saving compiled model to '{save_path}'")
        lammps_model_compiled.save(save_path)
        return save_path


def build_arg_parser():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Convert a franken model to be able to use it with LAMMPS",
    )
    parser.add_argument(
        "--model_path",
        type=str,
        help="Path to the model to be converted to LAMMPS",
    )
    parser.add_argument(
        "--rf_weight_id",
        type=int,
        help="Head of the model to be converted to LAMMPS",
        default=None,
    )
    return parser


def create_lammps_model_cli():
    parser = build_arg_parser()
    args = parser.parse_args()
    LammpsFrankenCalculator.create_lammps_model(args.model_path, args.rf_weight_id)  # type: ignore


if __name__ == "__main__":
    create_lammps_model_cli()


# For sphinx docs
get_parser_fn = lambda: build_arg_parser()  # noqa: E731



"""
Test the model conversion to LAMMPS (essentially testing torch-scriptability, not LAMMPS directly)
"""

import os
import pytest
import torch

from franken.backbones.wrappers.common_patches import unpatch_e3nn
from franken.config import GaussianRFConfig, MaceBackboneConfig, MultiscaleGaussianRFConfig
from franken.data import BaseAtomsDataset
from franken.rf.model import FrankenPotential
from franken.rf.scaler import Statistics
from franken.utils.misc import garbage_collection_cuda
from franken.datasets.registry import DATASET_REGISTRY
from franken.calculators.lammps_calc import LammpsFrankenCalculator

from .conftest import DEVICES
from .utils import are_dicts_close, cleanup_dir, create_temp_dir


RF_PARAMETRIZE = [
    GaussianRFConfig(num_random_features=128, length_scale=1.0),
    MultiscaleGaussianRFConfig(num_random_features=128),
]


@pytest.mark.parametrize("rf_cfg", RF_PARAMETRIZE)
@pytest.mark.parametrize("device", DEVICES)
def test_lammps_compile(rf_cfg, device):
    """Test for checking save and load methods of FrankenPotential"""
    unpatch_e3nn()  # needed in case some previous test ran the patching code
    gnn_cfg = MaceBackboneConfig("mace_mp/small")
    temp_dir = None
    try:
        # Step 1: Create a temporary directory for saving the model
        temp_dir = create_temp_dir()

        data_path = DATASET_REGISTRY.get_path("test", "test", None, False)
        dataset = BaseAtomsDataset.from_path(
            data_path=data_path,
            split="train",
            gnn_config=gnn_cfg,
        )
        model = FrankenPotential(
            gnn_config=gnn_cfg,
            rf_config=rf_cfg,
            scale_by_Z=True,
            num_species=dataset.num_species,
        ).to(device)

        with torch.no_grad():
            gnn_features_stats = Statistics()
            for data, _ in dataset:  # type: ignore
                data = data.to(device=device)
                gnn_features = model.gnn.descriptors(data)
                gnn_features_stats.update(
                    gnn_features, atomic_numbers=data.atomic_numbers
                )

            model.input_scaler.set_from_statistics(gnn_features_stats)
            garbage_collection_cuda()

        # Step 2: Save the model to the temporary directory
        model_save_path = os.path.join(temp_dir, "model_checkpoint.pth")
        model.save(model_save_path)

        # Step 3: Run create_lammps_model
        comp_model_path = LammpsFrankenCalculator.create_lammps_model(model_path=model_save_path, rf_weight_id=None)

        # Step 4: Load saved model
        comp_model = torch.jit.load(comp_model_path, map_location=device)

        # Step 4: Compare rf.state_dict between the original and loaded models
        with pytest.raises(RuntimeError) as exc:
            assert are_dicts_close(
                model.rf.state_dict(), comp_model.model.rf.state_dict(), verbose=True
            )
        assert "Float did not match Double" in str(exc.value)
        assert are_dicts_close(
            model.rf.double().state_dict(), comp_model.model.rf.state_dict(), verbose=True
        ), "The rf.state_dict() of the loaded model does not match the original model."

        with pytest.raises(RuntimeError) as exc:
            assert are_dicts_close(
                model.input_scaler.state_dict(),
                comp_model.model.input_scaler.state_dict(),
                verbose=True,
            )
        assert "Float did not match Double" in str(exc.value)
        assert are_dicts_close(
            model.input_scaler.double().state_dict(),
            comp_model.model.input_scaler.state_dict(),
            verbose=True,
        ), "The input_scaler.state_dict() of the loaded model does not match the original model."

        with pytest.raises(RuntimeError) as exc:
            assert are_dicts_close(
                model.energy_shift.state_dict(),
                comp_model.model.energy_shift.state_dict(),
                verbose=True,
            )
        assert "Float did not match Double" in str(exc.value)
        assert are_dicts_close(
            model.energy_shift.double().state_dict(),
            comp_model.model.energy_shift.state_dict(),
            verbose=True,
        ), "The energy_shift.state_dict() of the loaded model does not match the original model."
    finally:
        if temp_dir is not None:
            cleanup_dir(temp_dir)




from pathlib import Path
import shutil
import zipfile

import ase
import ase.io
import numpy as np

from franken.datasets.registry import DATASET_REGISTRY, BaseRegisteredDataset
from franken.utils.file_utils import download_file


@DATASET_REGISTRY.register("PtH2O")
class PtH2ORegisteredDataset(BaseRegisteredDataset):
    relative_paths = {
        "PtH2O": {
            "train": "PtH2O/train.extxyz",
            "val": "PtH2O/valid.extxyz",
            "test": "PtH2O/test.extxyz",
        },
    }

    @classmethod
    def get_path(
        cls, name: str, split: str, base_path: Path | None, download: bool = True
    ):
        if base_path is None:
            raise KeyError(None)
        relative_path = cls.relative_paths[name][split]
        path = base_path / relative_path
        if not path.is_file() and download:
            cls.download(base_path)
        if path.is_file():
            return path
        else:
            raise ValueError(f"Dataset not found at '{path.resolve()}'")

    @classmethod
    def download(cls, base_path: Path):
        pth2o_base_path = base_path / "PtH2O"
        pth2o_base_path.mkdir(exist_ok=True, parents=True)
        # Download
        download_file(
            url="https://data.dtu.dk/ndownloader/files/29141586",
            filename=pth2o_base_path / "data.zip",
            desc="Downloading PtH2O dataset",
            expected_md5="acd748f7f32c66961c90cb15457f7bae",
        )
        # Extract
        with zipfile.ZipFile(pth2o_base_path / "data.zip", "r") as zf:
            zf.extractall(pth2o_base_path)
        # Read full dataset
        full_traj = ase.io.read(
            pth2o_base_path / "Dataset_and_training_files" / "dataset.traj", index=":"
        )
        assert isinstance(full_traj, list)
        # Split into train/val/test
        np.random.seed(42)
        np.random.shuffle(full_traj)
        train_traj = full_traj[:30_000]
        valid_traj = full_traj[30_000:31_000]
        test_traj = full_traj[31_000:]
        # Saved shuffled to disk
        ase.io.write(pth2o_base_path / "train.extxyz", train_traj)
        ase.io.write(pth2o_base_path / "valid.extxyz", valid_traj)
        ase.io.write(pth2o_base_path / "test.extxyz", test_traj)
        # Cleanup
        (pth2o_base_path / "data.zip").unlink()
        shutil.rmtree(pth2o_base_path / "Dataset_and_training_files")


if __name__ == "__main__":
    PtH2ORegisteredDataset.download(Path(__file__).parent.parent)



"""Random feature implementations for different kernels"""

import dataclasses
from math import ceil, sqrt

import numpy as np
import scipy as sp
import torch

from franken.config import GaussianRFConfig, MultiscaleGaussianRFConfig, RFConfig
from franken.utils.misc import sanitize_init_dict

__all__ = (
    "RandomFeaturesHead",
    "OrthogonalRFF",
    "BiasedOrthogonalRFF",
    "MultiScaleOrthogonalRFF",
    "Linear",
    "TensorSketch",
    "initialize_rf",
)


class RandomFeaturesHead(torch.nn.Module):
    """Base class for random-feature heads

    Args:
        input_dim (int): Dimension of the input features.
        num_random_features (int): The number of random features to use in the feature mapping. Defaults to :math:`2^{10} = 1024`.
        num_species (int | None): The number of chemical species for
            which the kernel is computed. This parameter is relevant for systems
            with multiple chemical species. Defaults to :code:`None`.
        chemically_informed_ratio (float | None): The relative weight of chemically-informed kernels with respect to the all-species kernel. Ignored if
            :code:`num_species` is None. Defaults to :code:`None`.
    """

    def __init__(
        self,
        input_dim: int,
        num_random_features: int = 2**10,
        num_species: int | None = None,
        chemically_informed_ratio: float | None = None,
    ):
        super(RandomFeaturesHead, self).__init__()
        self.input_dim = input_dim
        self.num_random_features = num_random_features
        self.num_species = num_species
        self.chemically_informed_ratio = chemically_informed_ratio

        self._per_species_multiplier = 0
        if self.num_species is not None:
            assert self.num_species > 0
            self._per_species_multiplier += self.num_species
            if self.chemically_informed_ratio is not None:
                self._per_species_multiplier += 1
        else:
            self._per_species_multiplier += 1
        self.total_random_features = (
            self._per_species_multiplier * self.num_random_features
        )

        # Register weights buffer
        self.weights: torch.Tensor
        self.register_buffer("weights", torch.zeros((1, self.total_random_features)))

    def species_scatter_sum(
        self, Z: torch.Tensor, atomic_numbers: torch.Tensor | None = None
    ) -> torch.Tensor:
        r"""Average features across all atoms in a configuration.

        Depending on the configuration of random features, this function will either perform
        a simple average, or will use a chemically-informed averaging method where features
        are averaged within each atomic type and concatenated across atomic types. In this latter
        case, the number of output features is larger than the number of input features. It will
        always be equal to :code:`self.total_random_features`.

        Args:
            Z (torch.Tensor): [num_atoms, feature_dim] tensor containing the random features for each atom
            atomic_numbers (torch.Tensor | None): if specified, an integer tensor of size [num_atoms]
                detailing the atomic number of each atom in the configuration. Defaults to None.

        Returns:
            torch.Tensor: [total_feature_dim] tensor containing the random features for the whole configuration.
        """
        if self.num_species is None:
            return Z.mean(0)
        else:
            assert (
                atomic_numbers is not None
            ), "atomic_number should be specified when self.num_species is not None"
            species_map, scatter_idxs = torch.unique(
                atomic_numbers, sorted=True, return_inverse=True
            )
            assert (
                len(species_map) == self.num_species
            ), f"The provided atomic numbers {species_map}, are of a different number than self.num_species {self.num_species}"

            scatter_idxs = scatter_idxs.unsqueeze(-1).expand(
                Z.size()
            )  # ~[natoms, random_features_per_species] Broadcasting for backprop.
            chemically_informed_descriptors = (
                torch.zeros((self.num_species, self.num_random_features), dtype=Z.dtype)
                .to(Z.device)
                .scatter_reduce_(0, scatter_idxs, Z, "mean")
            ) / self.num_species  # Normalize.

            if self.chemically_informed_ratio is not None:
                kappa = self.chemically_informed_ratio
                assert (
                    0 <= kappa <= 1
                ), "The ratio of chemically informed feature feature map should be bounded between 0 and 1"
                chemically_informed_descriptors = torch.cat(
                    (
                        (1 - kappa) * Z.mean(0, keepdim=True),
                        kappa * chemically_informed_descriptors,
                    )
                )
            return chemically_informed_descriptors.view(-1)

    def init_args(self):
        """Returns the arguments needed to re-initialize this class."""
        return {
            "num_random_features": self.num_random_features,
            "num_species": self.num_species,
            "chemically_informed_ratio": self.chemically_informed_ratio,
        }


class OrthogonalRFF(RandomFeaturesHead):
    r"""
    `Orthogonal Random Fourier Features <https://arxiv.org/abs/1610.09072>`_ by Yu et al. for approximating the Gaussian kernel

    .. math::

        \text{exp}\left(-\frac{\| x - y \|^{2}}{2\ell^{2}}\right)

    As with the Gaussian kernel, the RF approximation depends on the :code:`length_scale` parameter.
    In addition to that, one needs to choose the number of random features to control the approximation
    quality.

    Args:
        input_dim (int): Dimensionality of the input features.
        num_random_features (int): The number of random features to use in the feature mapping. Defaults to :math:`2^{10} = 1024`.
        num_species (int | None): The number of chemical species for
            which the kernel is computed. This parameter is relevant for systems
            with multiple chemical species. Defaults to :code:`None`.
        chemically_informed_ratio (float | None): The relative weight of chemically-informed kernels with respect to the all-species kernel. Ignored if
            :code:`num_species` is None. Defaults to :code:`None`.
        use_offset (bool): A flag indicating whether to use an offset in
            the random feature generation. Using an offset reduces the number of
            random features by half but increases variance. Defaults to :code:`True`.
        length_scale (float): The length scale parameter :math:`\ell` that
            controls the smoothness of the kernel function. It affects how quickly
            the kernel values decay with distance. Defaults to 1.0.
        rng_seed (int | None): A seed for the random number generator
            used in generating random features. Setting this ensures reproducibility
            of results. Defaults to :code:`None`.
    """

    def __init__(
        self,
        input_dim: int,
        num_random_features: int = 2**10,
        num_species: int | None = None,
        chemically_informed_ratio: float | None = None,
        use_offset: bool = True,
        length_scale: float = 1.0,
        rng_seed=None,
    ):
        check_positive_arg(length_scale, "length_scale")
        super(OrthogonalRFF, self).__init__(
            input_dim,
            num_random_features,
            num_species=num_species,
            chemically_informed_ratio=chemically_informed_ratio,
        )
        self.rng_seed = rng_seed
        random_generator = get_seeded_generator(self.rng_seed)

        W = self._sample_orf_matrix(
            input_dim, self.num_random_features, random_generator, self.rng_seed
        )

        self.register_buffer("rff_matrix", W)
        self.register_buffer(
            "random_offset",
            torch.rand(self.num_random_features, generator=random_generator)
            * 2
            * np.pi,
        )
        self.register_buffer("length_scale", torch.tensor(length_scale))

        self.use_offset = use_offset
        if not self.use_offset:
            self.total_random_features = self.total_random_features * 2

    @staticmethod
    def _sample_orf_matrix(input_dim, num_random_features, torch_generator, rng_seed):
        num_folds = int(np.ceil(num_random_features / input_dim))

        G = torch.randn(
            num_folds,
            input_dim,
            input_dim,
            generator=torch_generator,
        )
        Q, _ = torch.linalg.qr(
            G, mode="complete"
        )  # The _columns_ in each batch of matrices in Q are orthonormal.

        Q = Q.transpose(
            -1, -2
        )  # The _rows_ in each batch of matrices in Q are orthonormal.

        S = torch.tensor(
            sp.stats.chi.rvs(
                input_dim,
                size=(num_folds, input_dim, 1),
                random_state=rng_seed,
            ),
        )

        W = Q * S  # [num_folds, input_dim, input_dim]
        W = torch.cat(
            [W[fold_idx, ...] for fold_idx in range(num_folds)], dim=0
        )  # Concatenate columns [num_heads, num_folds*input_dim, input_dim]
        W = W[:num_random_features, :]
        return W

    def feature_map(
        self,
        h: torch.Tensor,
        atomic_numbers: torch.Tensor | None = None,
    ) -> torch.Tensor:
        """Computes the random-feature map for a given configuration :code:`h`

        Args:
            h (torch.Tensor):
                descriptors for a single configuration ~[natoms, descriptors]
            atomic_numbers (torch.Tensor):
                atomic numbers for a single configuration ~[natoms]
        """
        length_scale = self.length_scale
        # Convert dtype of rf matrices to same as h
        self.rff_matrix = self.rff_matrix.to(dtype=h.dtype)
        if self.use_offset:
            self.random_offset = self.random_offset.to(dtype=h.dtype)
        # Compute the orthogonal random features nonlinear mapping
        Z = torch.einsum("af, Ff -> aF", h, self.rff_matrix / length_scale)
        if self.use_offset:
            b = self.random_offset.unsqueeze(0)
            Z = torch.cos(Z + b) / sqrt(0.5 * self.num_random_features)
        else:
            _s = torch.sin(Z)
            _c = torch.cos(Z)
            Z = torch.cat((_s, _c), dim=-1) / sqrt(self.num_random_features)

        return self.species_scatter_sum(
            Z, atomic_numbers=atomic_numbers
        )  # Sum over the single species

    def init_args(self):
        return super().init_args() | {
            "use_offset": self.use_offset,
            "length_scale": self.length_scale.tolist(),
            "rng_seed": self.rng_seed,
        }


class BiasedOrthogonalRFF(OrthogonalRFF):
    def __init__(
        self,
        input_dim: int,
        num_random_features: int = 2**10,
        num_species: int | None = None,
        chemically_informed_ratio: float | None = None,
        use_offset: bool = True,
        length_scale: float = 1.0,
        bias: float = 1.0,
        rng_seed=None,
    ):
        super(BiasedOrthogonalRFF, self).__init__(
            input_dim,
            num_random_features,
            num_species=num_species,
            chemically_informed_ratio=chemically_informed_ratio,
            use_offset=use_offset,
            length_scale=length_scale,
            rng_seed=rng_seed,
        )
        self.register_buffer("bias", torch.tensor(bias))

    def feature_map(
        self,
        h: torch.Tensor,
        atomic_numbers: torch.Tensor | None = None,
    ) -> torch.Tensor:
        self.bias = self.bias.to(dtype=h.dtype)
        # unbiased_features ~ [num_random_features - 1]
        features = super().feature_map(h, atomic_numbers=atomic_numbers)
        assert features.ndim == 1
        features[-1] = torch.sqrt(self.bias)
        return features

    def init_args(self):
        return super().init_args() | {"bias": self.bias.tolist()}


class MultiScaleOrthogonalRFF(RandomFeaturesHead):
    r"""
    A multi-scale version of :class:`OrthogonalRFF` which splits the available random features among multiple length-scales.
    This approximates a mixture of Gaussian kernels at different scales, simplifying hyper-parameter tuning.

    .. math::

        \text{exp}\left(-\frac{\| x - y \|^{2}}{2\ell^{2}}\right)

    Multiple scales are specified with arguments :code:`length_scale_low`, :code:`length_scale_high`
    and :code:`length_scale_num` which will be used to subdivide the available :code:`num_random_features`
    random features into :code:`length_scale_num` blocks with equally spaced length-scales.
    This means that each length-scale will only use a fraction of the total random features, but in practice
    we found this have very small impact on overall accuracy.
    This kernel can be seen as implicitly doing a grid-search over linearly spaced length-scales.

    Args:
        input_dim: Dimensionality of the input features.
        num_random_features: The number of random features to use in the feature mapping. Defaults to :math:`2^{10} = 1024`.
        num_species: The number of chemical species for
            which the kernel is computed. This parameter is relevant for systems
            with multiple chemical species. Defaults to :code:`None`.
        chemically_informed_ratio: The relative weight of chemically-informed kernels with respect to the all-species kernel. Ignored if
            :code:`num_species` is None. Defaults to :code:`None`.
        use_offset: A flag indicating whether to use an offset in
            the random feature generation. Using an offset reduces the number of
            random features by half but increases variance. Defaults to :code:`True`.
        length_scale_low: The lower end of the interval of length-scales considered.
        length_scale_high: The higher end of the interval of length-scales considered.
        length_scale_num: The number of different length-scales, equally spaced between
            :code:`length_scale_low` and :code:`length_scale_high` which are considered in the
            multi-scale approximation.
        rng_seed (int | None): A seed for the random number generator
            used in generating random features. Setting this ensures reproducibility
            of results. Defaults to :code:`None`.
    """

    def __init__(
        self,
        input_dim: int,
        num_random_features: int = 2**10,
        num_species: int | None = None,
        chemically_informed_ratio: float | None = None,
        use_offset: bool = True,
        length_scale_low: float = 1.0,
        length_scale_high: float = 10.0,
        length_scale_num: int = 4,
        rng_seed=None,
    ):
        super(MultiScaleOrthogonalRFF, self).__init__(
            input_dim,
            num_random_features,
            num_species=num_species,
            chemically_informed_ratio=chemically_informed_ratio,
        )
        self.rng_seed = rng_seed
        random_generator = get_seeded_generator(self.rng_seed)
        check_positive_arg(length_scale_low, "length_scale_low")
        self.length_scale_low = length_scale_low
        check_positive_arg(length_scale_high, "length_scale_high")
        self.length_scale_high = length_scale_high
        self.length_scale_num = length_scale_num

        W = OrthogonalRFF._sample_orf_matrix(
            input_dim, self.num_random_features, random_generator, self.rng_seed
        )

        self.register_buffer("rff_matrix", W)
        self.register_buffer(
            "random_offset",
            torch.rand(self.num_random_features, generator=random_generator)
            * 2
            * np.pi,
        )

        _num_repeats = ceil(self.num_random_features / self.length_scale_num)
        _length_scale = torch.repeat_interleave(
            torch.linspace(
                self.length_scale_low, self.length_scale_high, self.length_scale_num
            ),
            _num_repeats,
        )
        self.register_buffer("length_scale", _length_scale[: self.num_random_features])

        self.use_offset = use_offset
        if not self.use_offset:
            self.total_random_features = self.total_random_features * 2

    def feature_map(
        self,
        h: torch.Tensor,
        atomic_numbers: torch.Tensor | None = None,
    ) -> torch.Tensor:
        """Computes the random-feature map for a given configuration :code:`h`

        Args:
            h (torch.Tensor):
                descriptors for a single configuration ~[natoms, descriptors]
            atomic_numbers (torch.Tensor):
                atomic numbers for a single configuration ~[natoms]
        """
        # Convert dtype of rf matrices to same as h
        self.rff_matrix = self.rff_matrix.to(dtype=h.dtype)
        if self.use_offset:
            self.random_offset = self.random_offset.to(dtype=h.dtype)

        self.length_scale = self.length_scale.to(dtype=h.dtype)
        length_scale = self.length_scale.view(-1, 1)
        # Compute the orthogonal random features nonlinear mapping
        Z = torch.einsum("af, Ff -> aF", h, self.rff_matrix / length_scale)
        if self.use_offset:
            b = self.random_offset.unsqueeze(0)
            Z = torch.cos(Z + b) / sqrt(0.5 * self.num_random_features)
        else:
            _s = torch.sin(Z)
            _c = torch.cos(Z)
            Z = torch.cat((_s, _c), dim=-1) / sqrt(self.num_random_features)

        return self.species_scatter_sum(
            Z, atomic_numbers=atomic_numbers
        )  # Sum over the single species

    def init_args(self):
        return super().init_args() | {
            "use_offset": self.use_offset,
            "length_scale_low": self.length_scale_low,
            "length_scale_high": self.length_scale_high,
            "length_scale_num": self.length_scale_num,
            "rng_seed": self.rng_seed,
        }


class Linear(RandomFeaturesHead):
    r"""Implements the linear kernel.

    This kernel does not need to be approximated with random-features, and can be used directly
    as if it were composed of random features. No stochasticity is present.

    .. math::

        \frac{\langle x, y \rangle}{\ell^{2}} + b

    Args:
        input_dim (int): The dimensionality of the input features.
        num_species (int | None): The number of chemical species for
            which the kernel is computed. This parameter is relevant for systems
            with multiple chemical species. Defaults to :code:`None`.
        chemically_informed_ratio (float | None): The relative weight of chemically-informed kernels with respect to the all-species kernel. Ignored if
            :code:`num_species` is None. Defaults to :code:`None`.
        bias (float): The bias term :math:`b` added to the kernel function. This allows for shifting the kernel values, which can be useful for certain applications. Defaults to 0.0.
        length_scale (float): The length scale parameter :math:`\ell` that controls the smoothness of the kernel function. It affects how quickly the kernel values decay with distance. Defaults to 1.0.
    """

    def __init__(
        self,
        input_dim: int,
        num_species: int | None = None,
        chemically_informed_ratio: float | None = None,
        bias: float = 0.0,
        length_scale: float = 1.0,
    ):
        check_positive_arg(length_scale, "length_scale")
        super(Linear, self).__init__(
            input_dim,
            # Add a constant bias feature
            input_dim + 1,
            num_species=num_species,
            chemically_informed_ratio=chemically_informed_ratio,
        )

        self.register_buffer("bias", torch.tensor(bias))
        self.register_buffer("length_scale", torch.tensor(length_scale))

    def feature_map(
        self,
        h: torch.Tensor,
        atomic_numbers: torch.Tensor | None = None,
    ) -> torch.Tensor:
        """Computes the random-feature map for a given configuration :code:`h`

        Args:
            h (torch.Tensor):
                descriptors for a single configuration ~[natoms, descriptors]
            atomic_numbers (torch.Tensor):
                atomic numbers for a single configuration ~[natoms]
        """
        length_scale = self.length_scale
        scaled_descriptors = h / length_scale
        # Add a constant bias feature to scaled_descriptors
        bias_feature = (
            torch.sqrt(self.bias)
            .expand(scaled_descriptors.size(0), 1)
            .to(dtype=h.dtype)
        )
        scaled_descriptors = torch.cat([scaled_descriptors, bias_feature], dim=1)

        return self.species_scatter_sum(
            scaled_descriptors, atomic_numbers=atomic_numbers
        )

    def init_args(self):
        _super_init = super().init_args()
        _super_init.pop("num_random_features")
        return _super_init | {
            "bias": self.bias.tolist(),
            "length_scale": self.length_scale.tolist(),
        }


class TensorSketch(RandomFeaturesHead):
    r"""TensorSketch random-features approximation of the polynomial kernel

    .. math::
        \left(\frac{\langle x, y \rangle}{\ell^{2}} + b\right)^{d}


    from `"Fast and scalable polynomial kernels via explicit feature maps" <https://doi.org/10.1145/2487575.2487591>`_ by Pham and Pagh.

    Args:
        input_dim (int): The dimensionality of the input features.
        num_random_features (int): The number of random features to use in the sketching process. Defaults to :math:`2^{10} = 1024`.
        num_species (int | None): The number of chemical species for
            which the kernel is computed. This parameter is relevant for systems
            with multiple chemical species. Defaults to :code:`None`.
        chemically_informed_ratio (float | None): The relative weight of chemically-informed kernels with respect to the all-species kernel. Ignored if
            :code:`num_species` is None. Defaults to :code:`None`.
        degree (int): Degree :math:`d` of the polynomial kernel to sketch. Defaults to 2.
        bias (float): The bias term :math:`b` added to the kernel function. This allows for shifting the kernel values, which can be useful for certain applications. Defaults to 0.0.
        length_scale (float): The length scale parameter :math:`\ell` that controls the smoothness of the kernel function. It affects how quickly the kernel values decay with distance. Defaults to 1.0.
        rng_seed (int | None): A seed for the random number generator used in generating random features. Setting this ensures reproducibility of results. Defaults to :code:`None`.
    """

    def __init__(
        self,
        input_dim: int,
        num_random_features: int = 2**10,
        num_species: int | None = None,
        chemically_informed_ratio: float | None = None,
        degree=2,
        bias: float = 0.0,
        length_scale: float = 1.0,
        rng_seed=None,
    ):
        check_positive_arg(length_scale, "length_scale")
        super(TensorSketch, self).__init__(
            input_dim,
            num_random_features,
            num_species=num_species,
            chemically_informed_ratio=chemically_informed_ratio,
        )
        self.rng_seed = rng_seed
        self.register_buffer("degree", torch.tensor(degree))
        self.register_buffer("bias", torch.tensor(bias))
        self.register_buffer("length_scale", torch.tensor(length_scale))
        self.register_buffer(
            "count_sketches", self._sample_count_sketches(self.rng_seed, self.degree)
        )

    def _sample_count_sketches(self, rng_seed, degree: int):
        rng_torch = get_seeded_generator(rng_seed)
        with torch.no_grad():
            count_sketches = []
            for _ in range(degree):
                count_sketches.append(self._sample_hash_fns(rng_torch))
            return torch.stack(count_sketches)  # ~[degree, random_features, input_dim]

    def _sample_hash_fns(self, generator: torch.Generator | None):
        i_hash = torch.randint(
            low=0,
            high=self.num_random_features,
            size=(self.input_dim,),
            generator=generator,
        )
        s_hash = torch.bernoulli(
            0.5 * torch.ones(self.input_dim), generator=generator
        )  # Bernoulli
        # Convert to Rademacher
        s_hash = s_hash * 2 - 1
        indices = torch.stack([i_hash, torch.arange(self.input_dim)], dim=0)
        count_sketch = torch.sparse_coo_tensor(
            indices,
            s_hash,
            torch.Size([self.num_random_features, self.input_dim]),
        ).to_dense()
        return count_sketch

    def feature_map(
        self,
        h: torch.Tensor,
        atomic_numbers: torch.Tensor | None = None,
    ) -> torch.Tensor:
        """Computes the random-feature map for a given configuration :code:`h`

        Args:
            h (torch.Tensor):
                descriptors for a single configuration ~[natoms, descriptors]
            atomic_numbers (torch.Tensor):
                atomic numbers for a single configuration ~[natoms]
        """
        length_scale = self.length_scale
        scaled_descriptors = (h + torch.sqrt(torch.abs(self.bias))) / length_scale
        # Convert dtype of self to the same dtype as h.
        # Note: cannot use self.to(dtype) due to self.degree buffer being int!
        self.count_sketches = self.count_sketches.to(dtype=h.dtype)
        Z = torch.einsum(
            "af, dFf -> adF", scaled_descriptors, self.count_sketches
        )  # ~[atoms, degree, random_features]
        Z = torch.fft.rfft(Z, n=self.num_random_features)
        Z = Z.prod(1)  # ~[atoms, random_features]
        Z = torch.fft.irfft(Z, n=self.num_random_features)
        return self.species_scatter_sum(Z, atomic_numbers=atomic_numbers)

    def init_args(self):
        return super().init_args() | {
            "degree": self.degree.tolist(),
            "bias": self.bias.tolist(),
            "length_scale": self.length_scale.tolist(),
            "rng_seed": self.rng_seed,
        }


def check_positive_arg(value, name):
    if value <= 0:
        raise ValueError(f"Argument {name} must be positive. Got {value}.")


def get_seeded_generator(seed: int | None) -> torch.Generator | None:
    rng_torch = None
    if seed is not None:
        rng_torch = torch.Generator(device=torch.get_default_device())
        rng_torch.manual_seed(seed)
    return rng_torch


def initialize_rf(rf_config: RFConfig, rf_feature_dim: int):
    if isinstance(rf_config, GaussianRFConfig):
        rf_cls = OrthogonalRFF
    elif isinstance(rf_config, MultiscaleGaussianRFConfig):
        rf_cls = MultiScaleOrthogonalRFF
    else:
        raise ValueError(type(rf_config))
    random_features_params = sanitize_init_dict(rf_cls, dataclasses.asdict(rf_config))
    return rf_cls(input_dim=rf_feature_dim, **random_features_params)



import pytest
import torch


def test_registry_init():
    from franken.metrics.registry import registry

    assert hasattr(registry._instance, "_metrics")


def test_available_metrics():
    import franken.metrics as fm

    for name in ["energy_MAE", "forces_MAE", "forces_cosim"]:
        assert name in fm.available_metrics()


def test_register():
    import franken.metrics as fm
    from franken.metrics.base import BaseMetric

    class MockMetric(BaseMetric):
        pass

    assert "mock_metric" not in fm.available_metrics()
    fm.register("mock_metric", MockMetric)
    assert "mock_metric" in fm.available_metrics()


def test_init_metric():
    import franken.metrics as fm
    from franken.metrics.base import BaseMetric

    class MockMetric(BaseMetric):
        def __init__(self, device: torch.device, dtype: torch.dtype = torch.float32):
            super().__init__("mock_metric", device, dtype)

    fm.register("mock_metric", MockMetric)
    metric = fm.init_metric("mock_metric", torch.device("cpu"))
    assert isinstance(metric, MockMetric)


@pytest.mark.parametrize(
    (
        "metric_name",
        "targets_energy",
        "targets_forces",
        "pred_energy",
        "pred_forces",
        "atomic_numbers",
        "expected_scalar",
        "expected_entries",
        "expected_shape",
    ),
    [
        (
            "energy_MAE",
            torch.tensor(3.0),
            torch.tensor([[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]),
            torch.tensor(1.0),
            torch.tensor([[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]),
            None,
            1000.0,
            None,
            None,
        ),
        (
            "energy_RMSE",
            torch.tensor(3.0),
            torch.tensor([[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]),
            torch.tensor(1.0),
            torch.tensor([[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]),
            None,
            1000.0,
            None,
            None,
        ),
        (
            "forces_MAE",
            torch.tensor(0.0),
            torch.tensor([[1.0, 0.0, 0.0], [0.0, 2.0, 0.0]]),
            torch.tensor(0.0),
            torch.tensor([[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]),
            None,
            500.0,
            None,
            None,
        ),
        (
            "forces_RMSE",
            torch.tensor(0.0),
            torch.tensor([[1.0, 0.0, 0.0], [0.0, 2.0, 0.0]]),
            torch.tensor(0.0),
            torch.tensor([[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]),
            None,
            ((5.0 / 6.0) ** 0.5) * 1000.0,
            None,
            None,
        ),
        (
            "forces_RMSE2",
            torch.tensor(0.0),
            torch.tensor([[1.0, 0.0, 0.0], [0.0, 2.0, 0.0]]),
            torch.tensor(0.0),
            torch.tensor([[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]),
            None,
            ((5.0 / 6.0) ** 0.5) * 1000.0,
            None,
            None,
        ),
        (
            "forces_cosim",
            torch.tensor(0.0),
            torch.tensor([[1.0, 0.0, 0.0], [1.0, 0.0, 0.0]]),
            torch.tensor(0.0),
            torch.tensor([[1.0, 0.0, 0.0], [-1.0, 0.0, 0.0]]),
            None,
            0.0,
            None,
            None,
        ),
        (
            "forces_MAE_species",
            torch.tensor(0.0),
            torch.tensor([[1.0, 0.0, 0.0], [0.0, 2.0, 0.0], [0.0, 0.0, 3.0]]),
            torch.tensor(0.0),
            torch.tensor([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]),
            torch.tensor([1, 1, 8], dtype=torch.long),
            None,
            [
                (0, 0, 500.0),
                (0, 1, (1.0 / 3.0) * 1000.0),
                (0, 8, (2.0 / 3.0) * 1000.0),
            ],
            (1, 91),
        ),
        (
            "forces_RMSE_species",
            torch.tensor(0.0),
            torch.tensor([[1.0, 0.0, 0.0], [0.0, 2.0, 0.0], [0.0, 0.0, 3.0]]),
            torch.tensor([0.0, 0.0]),
            torch.tensor(
                [
                    [[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
                    [[1.0, 0.0, 0.0], [0.0, 2.0, 0.0], [0.0, 0.0, 3.0]],
                ]
            ),
            torch.tensor([1, 1, 8], dtype=torch.long),
            None,
            [
                (0, 1, ((1.0 / 3.0) ** 0.5) * 1000.0),
                (0, 8, ((4.0 / 3.0) ** 0.5) * 1000.0),
                (0, 0, (((1.0 / 3.0) ** 0.5 + (4.0 / 3.0) ** 0.5) / 2.0) * 1000.0),
                (1, 1, 0.0),
                (1, 8, 0.0),
                (1, 0, 0.0),
            ],
            (2, 91),
        ),
    ],
)
def test_metric_values_parametrized(
    metric_name,
    targets_energy,
    targets_forces,
    pred_energy,
    pred_forces,
    atomic_numbers,
    expected_scalar,
    expected_entries,
    expected_shape,
):
    import franken.metrics as fm
    from franken.data.base import Target

    metric = fm.init_metric(metric_name, torch.device("cpu"))
    targets = Target(energy=targets_energy, forces=targets_forces)
    predictions = Target(energy=pred_energy, forces=pred_forces)

    if atomic_numbers is None:
        metric.update(predictions, targets)
    else:
        assert metric.requires_species
        metric.update(predictions, targets, atomic_numbers=atomic_numbers)

    values = metric.compute()

    if expected_scalar is not None:
        assert values.item() == pytest.approx(expected_scalar)
    else:
        assert expected_shape is not None
        assert values.shape == expected_shape
        assert expected_entries is not None
        for model_idx, species_idx, expected in expected_entries:
            assert values[model_idx, species_idx].item() == pytest.approx(expected)


@pytest.mark.parametrize("metric_name", ["energy_MAE", "energy_RMSE"])
def test_energy_metrics_require_forces(metric_name):
    import franken.metrics as fm
    from franken.data.base import Target

    metric = fm.init_metric(metric_name, torch.device("cpu"))
    targets = Target(energy=torch.tensor(3.0), forces=None)
    predictions = Target(energy=torch.tensor(1.0), forces=None)
    with pytest.raises(NotImplementedError):
        metric.update(predictions, targets)



from typing import Callable, ClassVar
from pathlib import Path


class BaseRegisteredDataset:
    relative_paths: ClassVar[dict[str, dict[str, str]]]

    @classmethod
    def get_path(
        cls, name: str, split: str, base_path: Path | None, download: bool = True
    ) -> Path:
        raise NotImplementedError()

    @classmethod
    def is_valid_split(cls, name: str, split: str) -> bool:
        return split in cls.relative_paths[name]


_KT = str
_VT = type[BaseRegisteredDataset]


class DatasetRegistry(dict[_KT, _VT]):
    def register(self, name: _KT | list[_KT] | tuple[_KT]) -> Callable[[_VT], _VT]:
        def decorator(func: _VT) -> _VT:
            if isinstance(name, (list, tuple)):
                for name_single in name:
                    self[name_single] = func
            else:
                self[name] = func
            return func

        return decorator

    def get_path(
        self, name: str, split: str, base_path: Path | None, download: bool = True
    ):
        """Fetch the path for a dataset-split pair. If the dataset does not exist under
        the `base_path` directory, a download will be attempted.

        Args:
            name (str): dataset name (e.g. "water", "TM23/Ag-cold", "PtH2O")
            split (str): data-split, for example "train", "val" or "test"
            base_path (Path): the base path at which the dataset is stored.
            download (bool, optional): Whether to download the dataset if it does not exist.
                Defaults to True.

        Returns:
            dset_path (Path): a path to the ase-readable dataset.
        """
        return self[name].get_path(name, split, base_path, download)

    def is_valid_split(self, name: str, split: str) -> bool:
        return self[name].is_valid_split(name, split)


DATASET_REGISTRY = DatasetRegistry()



"""Franken model"""

import logging
import os
from typing import Literal, Mapping, Optional, Tuple, Union

import torch

from franken.config import (
    BackboneConfig,
    RFConfig,
)
from franken.backbones.utils import load_checkpoint
from franken.data import Configuration
from franken.rf.atomic_energies import AtomicEnergiesShift
from franken.rf.heads import initialize_rf
from franken.rf.scaler import FeatureScaler
from franken.utils.jac import jacfwd, tune_jacfwd_chunksize

logger = logging.getLogger("franken")


class FrankenPotential(torch.nn.Module):
    """Maps atomic configurations into high-dimensional vectors through a neural network and random features.

    This class provides both functions to map atom configurations with GNN + random features through methods
    :meth:`~franken.rf.model.FrankenPotential.feature_map` and :meth:`~franken.rf.model.FrankenPotential.grad_feature_map`,
    and to perform inference on the potential function of atom configurations given a learned
    linear model (:meth:`~franken.rf.model.FrankenPotential.energy_and_forces`).

    Args:
        gnn_config: Configuration object for the GNN. Sets the ID and other backbone parameters.
        rf_config: Configuration object for the random features (kernel-approximation). Sets all relevant kernel parameters,
            as well as the number of random features.
        jac_chunk_size: Force calculation requires computing Jacobians through the GNN. Since this is
            memory intensive, we can do this in batches across atoms of each configuration. When dealing with large systems,
            the chunk size becomes important to ensure no out-of-memory errors occur. This can either be set manually or be
            set automatically (by passing `"auto"`) which will attempt to determine the largest chunk that fits in memory.
            Defaults to "auto".
        scale_by_Z: Whether features should be scaled indipendently for each different species. Defaults to True.
        num_species: The number of species that this model will be trained on. Defaults to 1.
        atomic_energies: A precomputed dictionary mapping atomic numbers to the average energy of that species.
            Can be safely left to its default of None in most cases.

    Attributes:
        gnn (torch.nn.Module): The graph neural network which is used for feature extraction, loaded from
            a pretrained checkpoint.
        rf (torch.nn.Module): Random-features module :class:`franken.rf.heads.RandomFeaturesHead`.

    Note:
        The automatic Jacobian chunking is known to be error-prone. If you encounter out-of-memory errors
        with this option active, try manually setting the `jac_chunk_size` parameter.
    """

    def __init__(
        self,
        gnn_config: BackboneConfig,
        rf_config: RFConfig,
        jac_chunk_size: Union[int, Literal["auto"]] = "auto",
        scale_by_Z: bool = True,
        num_species: int = 1,
        atomic_energies: Optional[Mapping[int, torch.Tensor | float]] = None,
    ):
        super(FrankenPotential, self).__init__()
        # stored here as simpler to access than through self.gnn
        self.gnn_config = gnn_config
        self.rf_config = rf_config
        self.jac_chunk_size = jac_chunk_size
        self.num_species = num_species

        # Caches for jacobian function
        self._grad_energy_jacfn = None
        self._grad_fmap_jacfn = None

        # Initialize `gnn`, `rf`, `input_scaler`, `energy_shift` submodules
        self.gnn = load_checkpoint(gnn_config)

        rf_feature_dim = self.gnn.feature_dim()
        self.rf = initialize_rf(rf_config, rf_feature_dim=rf_feature_dim)

        self.input_scaler = FeatureScaler(
            input_dim=self.rf.input_dim,
            statistics=None,
            scale_by_Z=scale_by_Z,
            num_species=num_species,
        )
        self.energy_shift = AtomicEnergiesShift(
            num_species=num_species, atomic_energies=atomic_energies
        )

    @property
    @torch.jit.unused
    def hyperparameters(self):
        hps = {
            "franken": self.gnn_config.to_ckpt(),
            "random_features": self.rf_config.to_ckpt(),
            "input_scaler": self.input_scaler.init_args(),
        }
        return hps

    def save(self, path: os.PathLike | str, multi_weights: torch.Tensor | None = None):
        if multi_weights is not None:
            assert torch.is_tensor(multi_weights)
            assert multi_weights.ndim <= 2
            assert multi_weights.shape[-1] == self.rf.weights.shape[-1]

        ckpt = {
            "jac_chunk_size": self.jac_chunk_size,
            "multi_weights": multi_weights,
            "num_species": self.num_species,
        }

        ckpt = {
            "jac_chunk_size": self.jac_chunk_size,
            "multi_weights": multi_weights,
            "num_species": self.num_species,
            "rf": {
                "config": self.rf_config.to_ckpt(),
                "state_dict": self.rf.state_dict(),
            },
            "input_scaler": {
                "config": self.input_scaler.init_args(),
                "state_dict": self.input_scaler.state_dict(),
            },
            "energy_shift": self.energy_shift.state_dict(),
            "gnn": {
                "config": self.gnn_config.to_ckpt(),
            },
        }
        torch.save(ckpt, path)

    @classmethod
    def load(
        cls,
        path,
        map_location=None,
        rf_weight_id: int | None = None,
    ):
        ckpt = torch.load(path, map_location=map_location, weights_only=False)

        rf_cfg = RFConfig.from_ckpt(ckpt["rf"]["config"])
        gnn_cfg = BackboneConfig.from_ckpt(ckpt["gnn"]["config"])
        model = cls(
            gnn_config=gnn_cfg,
            rf_config=rf_cfg,
            jac_chunk_size=ckpt["jac_chunk_size"],
            num_species=ckpt["num_species"],
            **ckpt["input_scaler"]["config"],
        )
        model.rf.load_state_dict(ckpt["rf"]["state_dict"])
        model.input_scaler.load_state_dict(ckpt["input_scaler"]["state_dict"])
        model.energy_shift.load_state_dict(ckpt["energy_shift"])

        if ckpt["multi_weights"] is not None:
            if rf_weight_id is None:
                raise ValueError(
                    f"The checkpoint contains {ckpt['multi_weights'].shape[0]}, select which one to load by specifying rf_weight_id"
                )
            assert rf_weight_id < ckpt["multi_weights"].shape[0]
            model.rf.weights.copy_(
                ckpt["multi_weights"][rf_weight_id].reshape_as(model.rf.weights)
            )

        if map_location is not None:
            return model.to(map_location)
        else:
            return model

    def feature_map(self, data: Configuration):
        """Obtain an embedding of each atom, and map it through
        random features. The RF mapping computes an average so
        the final feature map is per-configuration, instead of
        per-atom.
        """
        gnn_descriptors = self.gnn.descriptors(data)

        normalized_descriptors = self.input_scaler(
            gnn_descriptors,
            atomic_numbers=data.atomic_numbers,
        )
        return self.rf.feature_map(
            normalized_descriptors,
            atomic_numbers=data.atomic_numbers,
        )

    def _feature_map_aux(self, atom_pos: torch.Tensor, data: Configuration):
        old_atom_pos = data.atom_pos
        data.atom_pos = atom_pos
        random_features = self.feature_map(data)
        data.atom_pos = old_atom_pos
        return random_features, random_features

    def grad_feature_map(
        self, data: Configuration
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Compute the feature map for this configuration and
        its gradient with respect to atomic positions

        Returns:
         - forces_fmap : Tensor of size [n_random_features, n_atoms * 3]
         - energy_fmap : Tensor of size [n_random_features]
        """
        if self._grad_fmap_jacfn is None:
            jac_chunk_size = self.get_jacobian_chunk_size(
                self._feature_map_aux, [data.atom_pos, data], argnums=0
            )
            self._grad_fmap_jacfn = jacfwd(
                self._feature_map_aux,
                argnums=0,
                has_aux=True,
                chunk_size=jac_chunk_size,
            )
        return self._grad_fmap_jacfn(data.atom_pos, data)

    def energy(
        self, weights: Optional[torch.Tensor], data: Configuration
    ) -> torch.Tensor:
        r"""Computes the energy of a configuration, using a linear model.

        if `weights` is not provided, the weights stored in the :attr:`FrankenPotential.rf`
        random features object will be used instead.

        This function returns the energy of the configuration scaled by the number
        of atoms in the configuration itself

        .. math::
            \text{out} = \text{num atoms} \times E(\text{configuration})

        Args:
            weights: The linear coefficients of the energy model.
            configuration: The molecular configuration whose energy to compute.
        """
        if weights is None:
            weights = self.rf.weights

        feature_map = self.feature_map(data).to(dtype=weights.dtype)

        # Contract the last dim of weights with the first of feature_map
        energy = data.natoms * torch.tensordot(weights, feature_map, dims=1)

        return energy

    def _energy_aux(
        self,
        weights: Optional[torch.Tensor],
        atom_pos: torch.Tensor,
        data: Configuration,
    ):
        old_atom_pos = data.atom_pos
        data.atom_pos = atom_pos
        energy = self.energy(weights, data)
        data.atom_pos = old_atom_pos
        return energy, energy

    def grad_energy_func(
        self, weights: Optional[torch.Tensor], data: Configuration
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Computes the gradient of the :meth:`~franken.rf.model.FrankenPotential.energy` acting on a configuration.

        The gradient is equivalent to the negative force acting on
        the configuration.

        if `weights` is not provided, the weights stored in the :attr:`FrankenPotential.rf`
        random features object will be used instead.

        This function returns a tuple: `energy_gradient`, `energy`.

        .. note::
            This function uses the :mod:`torch.func` package to compute the gradient,
            which is particularly useful when computing the energy gradient with
            multiple linear models. In this case `weights` can be a matrix whose
            first dimension is the number of linear models.
            When computing the gradient for a single linear model, use the
            :meth:`~franken.rf.model.FrankenPotential.grad_energy_autograd` method instead
            for better performance.

        Args:
            weights (Tensor or None): the linear coefficients to compute the energy
            data: the molecular configuration whose energy and gradient to compute.

        See also :meth:`~franken.rf.model.FrankenPotential.grad_energy_autograd`.

        """
        if self._grad_energy_jacfn is None:
            jac_chunk_size = self.get_jacobian_chunk_size(
                self._energy_aux, [weights, data.atom_pos, data], argnums=1
            )
            self._grad_energy_jacfn = jacfwd(
                self._energy_aux, argnums=1, has_aux=True, chunk_size=jac_chunk_size
            )
        out = self._grad_energy_jacfn(weights, data.atom_pos, data)
        return out

    def grad_energy_autograd(
        self, weights: Optional[torch.Tensor], data: Configuration
    ) -> Tuple[torch.Tensor | None, torch.Tensor]:
        """Computes the gradient of the :meth:`~franken.rf.model.FrankenPotential.energy` acting on a configuration.

        The gradient is equivalent to the negative force acting on
        the configuration.

        if `weights` is not provided, the weights stored in the :attr:`FrankenPotential.rf`
        random features object will be used instead.

        This function returns a tuple: `energy_gradient`, `energy`. See
        :meth:`~franken.rf.model.FrankenPotential.grad_energy_func` for a discussion on the
        performance characteristics of the two implementations.

        Args:
            weights (Tensor or None): the linear coefficients to compute the energy
            data: the molecular configuration whose energy and gradient to compute.
        """
        # Ensure atom positions require gradients
        data.atom_pos.requires_grad_(True)
        # Compute the energy
        energy = self.energy(weights, data)

        if energy.ndim == 0:
            # Scalar case, single gradient
            gradient = torch.autograd.grad(outputs=[energy], inputs=[data.atom_pos])[0]
        else:
            # Vector case, compute gradients for each element independently
            gradients: list[torch.Tensor] = []
            for i in range(energy.shape[0]):
                grad_i = torch.autograd.grad(
                    outputs=[energy[i]],
                    inputs=[data.atom_pos],
                    retain_graph=True,
                )[0]
                assert grad_i is not None
                gradients.append(grad_i)
            # Stack gradients along a new dimension
            gradient = torch.stack(gradients, dim=0)

        return gradient, energy

    def get_jacobian_chunk_size(self, func, func_inputs, argnums=0) -> int:
        if hasattr(self, "_auto_jac_chunk_size"):
            return self._auto_jac_chunk_size
        else:
            jac_chunk_size = self.jac_chunk_size
            if jac_chunk_size == "auto":
                # TODO: We can probably cache the value for different functions to avoid multiple tuner runs.
                jac_chunk_size = tune_jacfwd_chunksize(
                    test_sample=func_inputs,
                    func=func,
                    argnums=argnums,
                    has_aux=True,
                )
                self._auto_jac_chunk_size = jac_chunk_size
                logger.info(
                    f"jacobian chunk size automatically set to {self._auto_jac_chunk_size}"
                )
            assert isinstance(jac_chunk_size, int)
            return jac_chunk_size

    def energy_and_forces(
        self,
        data: Configuration,
        weights: Optional[torch.Tensor] = None,
        forces_mode: str = "torch.autograd",
        add_energy_shift: bool = True,
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        """Infer energy and forces of an atomic configuration using a learned random-features model.

        The parameter `weights` can be used to specified the model's weights. Otherwise the weights stored in
        :attr:`FrankenPotential.rf.weights` will be used instead.

        The different values of `forces_mode` correspond to different ways of differentiating
        through the model to obtain the forces acting on the atoms:

        * :code:`"torch.func"` is best for when `weights` contains multiple linear models on which to perform inference at the same time (in that case `weights` should be a matrix of shape `[n_linear_models, model_size]`).

        * :code:`"torch.autograd"` is best for when a single linear model is used (i.e. when `weights` is a vector of shape `[model_size]`)

        * :code:`"no_forces"` can be used if forces are not required.

        Args:
            weights: weights of the random feature model. Defaults to None, in which case the weights set in :attr:`FrankenPotential.rf` will be used instead.

            forces_mode: how to compute the model's forces. Defaults to :code:`"torch.autograd"`.

            add_energy_shift: whether to add the energy shift to the energy.

        Returns:
            A tuple containing a tensor representing the potential energy (this is scalar, unless doing inference with multiple
            models when it can be vector-valued), and another optional tensor representing the forces acting on each atom of the
            given configuration. If multiple models are given, the forces will have shape :code:`[n_linear_models, n_atoms, 3]`,
            otherwise they will have shape :code:`[n_atoms, 3]`.
        """
        if forces_mode == "torch.func":
            with torch.no_grad():
                grad_energy, energy = self.grad_energy_func(weights, data)
                forces = -grad_energy.detach()
        elif forces_mode == "torch.autograd":
            grad_energy, energy = self.grad_energy_autograd(weights, data)
            assert grad_energy is not None
            forces = -grad_energy.detach()
        elif forces_mode == "no_forces":
            energy = self.energy(weights, data)
            forces = None
        else:
            raise ValueError(f"forces_mode '{forces_mode}' is not valid.")

        if add_energy_shift:
            energy = energy + self.energy_shift(data.atomic_numbers)
        return energy.detach(), forces

    def energy_and_forces_from_fmaps(
        self,
        data: Configuration,
        energy_fmap: torch.Tensor,
        forces_fmap: torch.Tensor,
        weights: Optional[torch.Tensor] = None,
        add_energy_shift: bool = True,
    ):
        if weights is None:
            weights = self.rf.weights
        energy = torch.tensordot(
            weights,
            data.natoms * energy_fmap,
            dims=([1], [0]),  # type: ignore
        )
        forces = torch.tensordot(
            weights,
            data.natoms * forces_fmap.view(forces_fmap.shape[0], -1, 3),
            dims=([1], [0]),  # type: ignore
        )
        if add_energy_shift:
            energy = energy + self.energy_shift(data.atomic_numbers)
        return energy, forces

    def forward(
        self,
        data: Configuration,
        weights: Optional[torch.Tensor] = None,
        add_energy_shift: bool = True,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        See docstring of :meth:`~franken.rf.model.FrankenPotential.energy_and_forces`.

        This function defaults to using the 'torch.autograd' strategy which allows the model
        to be jit-compiled.
        """
        grad_energy, energy = self.grad_energy_autograd(weights, data)
        assert grad_energy is not None
        forces = -grad_energy
        if add_energy_shift:
            energy = energy + self.energy_shift(data.atomic_numbers)
        return energy.detach(), forces.detach()



from franken.datasets.registry import DATASET_REGISTRY

# Ensure all sub-datasets are imported so that they are registered.
from .water import water_dataset  # noqa: F401
from .TM23 import tm23_dataset  # noqa: F401
from .PtH2O import pth2o_dataset  # noqa: F401
from .test import test_dataset  # noqa: F401

__all__ = ("DATASET_REGISTRY",)
